(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import predicate.

Require Import words.
Require Import constraints.
Require Import pme.

Require Import bbi_simple.
 
Section Elem.

  Set Implicit Arguments.

  Variable L : Type.

  Inductive bbi_basic (R : constraint L -> Prop) : constraint L -> Prop :=
    | In_bbi_basic_1 : forall a b m, R (m ~~ m) -> a<> b -> ~ alphabet R a -> ~ alphabet R b -> bbi_basic R (l a o l b ~~ m)
    | In_bbi_basic_2 : forall a b m, R (m ~~ m) -> a<> b -> ~ alphabet R a -> ~ alphabet R b -> bbi_basic R (l a o m ~~ l b)
    | In_bbi_basic_3 : forall m, R (m ~~ m) -> bbi_basic R (eps ~~ m).

  Proposition bbi_basic_id R R' c : R ~ R' -> bbi_basic R c -> bbi_basic R' c.
  Proof.
    intros H1 [ ]; [ constructor 1 | constructor 2 | constructor 3 ]; auto; try (apply H1; auto);
    match goal with [ H: ~ alphabet _ ?x |- ~ alphabet _ ?x ] =>  contradict H; generalize H; apply alphabet_inc, H1 end.
  Qed.
    
  Proposition bbi_basic_eps R : is_pme R -> bbi_basic R (eps~~eps).
  Proof. constructor 3; apply ceq_eps with (1 := H). Qed.

  Definition bbi_basic_pme := basic_pme bbi_basic.
  Definition bbi_simple_pme := simple_pme bbi_basic.

  Hint Resolve bbi_basic_id bbi_basic_eps.

  Proposition bbi_basic_inc_simple : bbi_basic_pme inc bbi_simple_pme.
  Proof. apply basic_inc_simple_pme; auto. apply bbi_basic_id. Qed.

  Proposition bbi_simple_inc_is_pme : bbi_simple_pme inc (@is_pme _).
  Proof. apply simple_is_pme. Qed.
    
End Elem.
