(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import predicate.
Require Import closure.
Require Import carrier. 
Require Import finite.

Require Import labels.
Require Import formula.
Require Import words.
Require Import constraints.
Require Import pme.
Require Import css.

Require Import branch.
Require Import bbi_expansion.

Inductive bbi_closed_branch_cond : CSS_type -> Prop :=
  | in_bbi_cb_0 : forall (X : _->Prop) C A m n, X (T A # m) -> X (F A # n) -> pme C (m ~~ n) -> bbi_closed_branch_cond (X,C)
  | in_bbi_cb_1 : forall (X : _->Prop) C m, X (F I # m) -> pme C (eps~~m) -> bbi_closed_branch_cond (X,C).

Definition bbi_closed_branch := CSS cap bbi_closed_branch_cond.

Hint Extern 2 ( bbi_closed_branch _ ) => apply bin_inter_sg'. 

Section bbi_cb_inc.

  Let bbi_cb_inc_cond c1 c2 : fst c1 inc fst c2 -> pme (snd c1) inc pme (snd c2) -> bbi_closed_branch_cond c1 -> bbi_closed_branch_cond c2.
  Proof.
    intros HX HC H.  
    induction H as [ X C A m n H1 H2 H3 | X C m H1 H2 ]; destruct c2 as (X',C'); simpl in HX, HC.
    
    apply HX in H1.
    apply HX in H2.
    apply HC in H3.
    constructor 1 with (1 := H1) (3 := H3); auto.
    
    apply HX in H1.
    apply HC in H2.
    constructor 2 with (1 := H1); auto.
  Qed.
      
  Proposition bbi_cb_inc X C X' C' : X inc X' -> pme C inc pme C' -> CSS (X',C') -> bbi_closed_branch (X,C) -> bbi_closed_branch (X',C').
  Proof.
    intros H1 H2 H3 H4.
    apply bin_inter_sg in H4.
    destruct H4 as [ H4 H5 ].
    apply bin_inter_sg'; auto.
    apply bbi_cb_inc_cond with (3 := H5); auto.
  Qed.

End bbi_cb_inc.

Proposition bbi_cb_inc_carrier : forall l1 l2, listbranch_eq CSS_inc l1 l2 -> carrier l1 inc bbi_closed_branch -> carrier l2 inc bbi_closed_branch.
Proof.
  apply listbranch_eq_lift.
  intros [] [] ( H1 & H2 & H3 ) H.
  generalize H.
  apply bbi_cb_inc ; auto.
  apply H1.
  apply bin_inter_sg in H.
  intros s f m Hf. 
  apply H1 in Hf.  
  apply H2, (proj1 H) with (1 := Hf).
Qed.
  
Proposition bbi_cb_subst sigma css : bbi_closed_branch css -> bbi_closed_branch (subst_CSS sigma css).
Proof.
  intros H.
  apply bin_inter_sg in H.
  destruct H as [ Hcss H2 ].
  apply bin_inter_sg'.
  destruct css; apply CSS_subst_stable; auto.
  induction H2 as [ X C A m n H1 H2 H3 | X C m H1 H2 ].
  
  apply subst_pme_prop with (sigma := sigma) in H3.
  constructor 1 with (A := A) (3 := H3).
  exists (T A # m); auto.
  exists (F A # n); auto.
  
  apply subst_pme_prop with (sigma := sigma) in H2.
  simpl in H2.
  constructor 2 with (2 := H2).
  exists (F I # m); auto.
Qed.

Hint Resolve bbi_cb_subst.

Proposition bbi_cb_subst_carrier sigma : forall lcss, carrier lcss                         inc bbi_closed_branch 
                                                   -> carrier (map (subst_CSS sigma) lcss) inc bbi_closed_branch.
Proof. apply list_lift; auto. Qed.

Proposition bbi_cb_expand : forall css1 css2, CSS css1 -> bbi_closed_branch css2 -> bbi_closed_branch (bbi_expand_1 css1 css2).
Proof.
   intros (X1,C1) (X,C) H1 HXC.
   apply bbi_cb_inc with (4 := HXC); auto.
   apply bin_inter_sg in HXC; destruct HXC; auto.
Qed.

Hint Resolve bbi_cb_expand.

Proposition bbi_cb_expand_carrier css : CSS css -> forall lcss, carrier lcss inc bbi_closed_branch -> carrier (bbi_expand css lcss) inc bbi_closed_branch.
Proof. intro; apply list_lift; auto. Qed.

Proposition bbi_cb_finite css : bbi_closed_branch css -> exists cssf, bbi_closed_branch cssf /\ CSS_leqf cssf css.
Proof.
  intros H.
  apply bin_inter_sg in H.
  destruct H as [ Hcss H2 ].
  induction H2 as [ X C A m n H1 H2 H3 | X C m H1 H2 ].
  destruct (pme_cpct H3) as ( Cf & H4 & H5 & H6 ).
  assert (pme Cf (m ~~ m)) as Hm.
    apply ceq_lft with (2 := H6); auto.
  assert (pme Cf (n ~~ n)) as Hn.
    apply ceq_rt  with (2 := H6); auto.
  exists (sg (T A # m) cup sg (F A # n), Cf).
  split; auto.
  apply bin_inter_sg'.
  intros s f x Hx; apply bin_union_sg in Hx.
  destruct Hx as [ Hx | Hx ]; apply in_sg_eq in Hx; injection Hx; clear Hx; intros Hx _ _ ; rewrite <- Hx; auto.
  constructor 1 with (A := A) (m := m) (n := n); apply in_sg; auto.

  destruct (pme_cpct H2) as ( Cf & H4 & H5 & H6 ).
  assert (pme Cf (m ~~ m)) as Hm.
    apply ceq_rt with (2 := H6); auto.
  exists (sg (F I # m), Cf).
  split; auto.
  apply bin_inter_sg'.
  intros s f x Hx; apply in_sg_eq in Hx; injection Hx; clear Hx; intros Hx _ _ ; rewrite <- Hx; auto.
  constructor 2 with (m := m); apply in_sg; auto.
Qed.


  