(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import predicate.

Definition injective {X Y} (f : X -> Y) := forall x x', f x = f x' -> x = x'. 
Definition surjective {X Y} (f : X -> Y) := forall y, exists x, y = f x.

Definition d_image {X Y} (f : X -> Y) y : Prop := exists x, f x = y.
Definition r_image {X Y} (f : X -> Y) y x : Prop := f x = y. 

Definition pd_image {X Y} (f : X -> Y) (P : X -> Prop) y : Prop := exists x, P x /\ f x = y.
Definition pr_image {X Y} (f : X -> Y) (Q : Y -> Prop) x : Prop := exists y, Q y /\ f x = y. 

Definition bounded (P : nat -> Prop) := exists k, forall n, P n -> n <= k.
Definition unbounded (P : nat -> Prop) := forall n, exists k, n <= k /\ P k.
Definition i_surjective {X} (f : nat -> X) := forall x, unbounded (r_image f x). 

Proposition pd_image_trivial X Y (f : X -> Y) (P : _ -> Prop) x : P x -> pd_image f P (f x).
Proof. exists x; auto. Qed.

Hint Resolve pd_image_trivial.

Proposition pd_image_empty X Y (f : X -> Y) : pd_image f empty ~ empty.
Proof.
  split; auto. 
  intros ? [ ? [ Hx _ ] ]; contradiction Hx.
Qed.

Proposition pd_image_sg X Y (f : X -> Y) x : pd_image f (sg x) ~ sg (f x).
Proof.
  split; intros y. 
  intros [ k [ Hx Hk ] ]; pred_solve; rewrite Hx, Hk; auto. 
  intros Hy; exists x; pred_solve.
Qed.

Hint Immediate pd_image_empty pd_image_sg.

Proposition pd_image_inc X Y (f : X -> Y) P Q : P inc Q -> pd_image f P inc pd_image f Q.
Proof. intros ? ? [ y [ ? ? ] ]; exists y; auto. Qed.

Proposition pr_image_inc X Y (f : X -> Y) P Q : P inc Q -> pr_image f P inc pr_image f Q.
Proof.  intros ? ? [ y [ ? ? ] ]; exists y; auto. Qed.

Hint Resolve pd_image_inc pr_image_inc.

Proposition prd_image_inc X Y (f : X -> Y) P : P inc pr_image f (pd_image f P).
Proof. intros x ?. exists (f x); split; auto. Qed.

Hint Immediate prd_image_inc.
  
Proposition pd_image_id X Y (f : X -> Y) P Q : P ~ Q -> pd_image f P ~ pd_image f Q.
Proof. intros [ ]; split; auto. Qed.

Hint Resolve pd_image_id.

Proposition pd_image_inc_cup X Y (f : X -> Y) P Q : pd_image f (P cup Q) inc pd_image f P cup pd_image f Q.
Proof.
  intros y Hy.
  destruct Hy as [ x [ Hx Hf ] ].
  pred_solve; [ apply bin_union_inc_r1 | apply bin_union_inc_r2 ]; exists x; tauto.
Qed.

Hint Immediate pd_image_inc_cup.

Proposition pd_image_id_cup X Y (f : X -> Y) P Q : pd_image f (P cup Q) ~ pd_image f P cup pd_image f Q.
Proof. split; auto. Qed.

Proposition pd_image_identity X (f : X -> X) P : (forall x, f x = x) -> pd_image f P ~ P.
Proof.
  intros Hf. 
  split.
  intros y [ x [ Hx1 Hx2 ] ].
  rewrite <- Hx2, Hf; auto.
  intros y Hy.
  exists y; auto.
Qed.

Proposition isurj_imp_surj X (f : nat -> X) : i_surjective f -> surjective f.
Proof.
  intros Hf x.
  destruct (Hf x 0) as [k [ _ Hk ]].
  exists k; auto.
Qed.

Proposition comp_surj X Y Z (f : X -> Y) (g : Y -> Z) : surjective f -> surjective g -> surjective (fun x => g (f x)).
Proof.
  intros Hf Hg z.
  destruct (Hg z) as [ y Hy ].
  destruct (Hf y) as [ x Hx ].
  exists x; rewrite <- Hx; auto.
Qed.

Proposition comp_isurj_surj X Y (f : nat -> X) (g : X -> Y) : i_surjective f -> surjective g -> i_surjective (fun x => g (f x)).
Proof.
  intros Hf Hg y n.
  destruct (Hg y) as [x Hx].
  destruct (Hf x n) as [k Hk].
  destruct Hk as [Hk Hk'].
  exists k. split. trivial.
  red in Hk' |- *.
  rewrite -> Hk'; auto.
Qed.